diff --git a/src/backend/access/hash/hashfunc.c b/src/backend/access/hash/hashfunc.c
index 9ee654e..5a00a03 100644
--- a/src/backend/access/hash/hashfunc.c
+++ b/src/backend/access/hash/hashfunc.c
@@ -297,6 +297,9 @@ hashvarlena(PG_FUNCTION_ARGS)
  * of 2.  There is no need to do mod a prime (mod is sooo slow!).
  * If you need less than 32 bits, use a bitmask.
  *
+ * This procedure never fails. Its important. Some code notably ResourceOwner
+ * relies on this.
+ *
  * Note: we could easily change this function to return a 64-bit hash value
  * by using the final values of both b and c.  b is perhaps a little less
  * well mixed than c, however.
diff --git a/src/backend/utils/resowner/resowner.c b/src/backend/utils/resowner/resowner.c
index 3330c8d..abed36c 100644
--- a/src/backend/utils/resowner/resowner.c
+++ b/src/backend/utils/resowner/resowner.c
@@ -37,29 +37,60 @@
  * which should be called before corresponding ResourceOwnerRemember* calls
  * (see below). Internally each type of resource is stored in separate
  * ResourceArray.
+ *
+ * There are two major reasons for using ResourceArray instead of, say,
+ * regular C arrays.
+ *
+ * Firstly we would like to prevent code duplication. For instance
+ * ResourceArray provides generic Remember/Forget/Enlarge procedures, so
+ * corresponding ResourceOwner* procedures are just a typesafe wrappers for
+ * these procedures.
+ *
+ * Secondly ResourceArray must be more efficient than regular C array.
+ * Current implementation in general could be considered a hash table. It has
+ * O(1) complexity of both Remember and Forget procedures.
  */
 typedef struct ResourceArray
 {
 	Datum	   *itemsarr;		/* buffer for storing values */
+	Datum		invalidval;		/* value that is considered invalid */
 	uint32		capacity;		/* capacity of array */
 	uint32		nitems;			/* how many items is stored in items array */
+	uint32		maxitems;		/* precalculated RESARRAY_MAX_ITEMS(capacity) */
+	uint32		lastidx;		/* index of last item returned by GetAny */
 }	ResourceArray;
 
 /*
  * This number is used as initial size of resource array. If given number of
  * items is not enough, we double array size and reallocate memory.
+ *
+ * Should be power of two since we use (arrsize - 1) as mask for hash value.
+ *
  */
 #define RESARRAY_INIT_SIZE 16
 
 /*
+ * How many items could be stored in a resource array of given capacity. If
+ * this number is reached we need to resize an array to prevent hash collisions.
+ *
+ * This computation actually costs only two additions and one binary shift.
+ */
+#define RESARRAY_MAX_ITEMS(capacity) ((capacity)*3/4)
+
+/*
  * Initialize ResourceArray
  */
 static void
-ResourceArrayInit(ResourceArray * resarr)
+ResourceArrayInit(ResourceArray * resarr, Datum invalidval)
 {
 	Assert(resarr->itemsarr == NULL);
 	Assert(resarr->capacity == 0);
 	Assert(resarr->nitems == 0);
+	Assert(resarr->maxitems == 0);
+	Assert(resarr->invalidval == 0);
+	Assert(resarr->lastidx == 0);
+
+	resarr->invalidval = invalidval;
 }
 
 /*
@@ -70,11 +101,24 @@ ResourceArrayInit(ResourceArray * resarr)
 static void
 ResourceArrayAdd(ResourceArray * resarr, Datum data)
 {
+	Datum		idx;
+	Datum		mask = resarr->capacity - 1;
+
+	Assert(resarr->maxitems > resarr->nitems);
 	Assert(resarr->capacity > 0);
 	Assert(resarr->itemsarr != NULL);
-	Assert(resarr->nitems < resarr->capacity);
+	Assert(data != resarr->invalidval);
+
+	idx = hash_any((void *) &data, sizeof(data)) & mask;
+
+	while (true)
+	{
+		if (resarr->itemsarr[idx] == resarr->invalidval)
+			break;
+		idx = (idx + 1) & mask;
+	}
 
-	resarr->itemsarr[resarr->nitems] = data;
+	resarr->itemsarr[idx] = data;
 	resarr->nitems++;
 }
 
@@ -86,24 +130,24 @@ ResourceArrayAdd(ResourceArray * resarr, Datum data)
 static bool
 ResourceArrayRemove(ResourceArray * resarr, Datum data)
 {
-	int			i,
-				j,
-				lastidx;
+	uint32		i;
+	Datum		idx;
+	Datum		mask = resarr->capacity - 1;
 
 	Assert(resarr->capacity > 0);
 	Assert(resarr->itemsarr != NULL);
+	Assert(data != resarr->invalidval);
 
-	lastidx = ((int) resarr->nitems) - 1;
-
-	for (i = lastidx; i >= 0; i--)
+	idx = hash_any((void *) &data, sizeof(data)) & mask;
+	for (i = 0; i < resarr->capacity; i++)
 	{
-		if (resarr->itemsarr[i] == data)
+		if (resarr->itemsarr[idx] == data)
 		{
-			for (j = i; j < lastidx; j++)
-				resarr->itemsarr[j] = resarr->itemsarr[j + 1];
+			resarr->itemsarr[idx] = resarr->invalidval;
 			resarr->nitems--;
 			return true;
 		}
+		idx = (idx + 1) & mask;
 	}
 
 	return false;
@@ -119,27 +163,33 @@ static void
 ResourceArrayEnlarge(ResourceArray * resarr)
 {
 	uint32		i,
-				oldcap,
-				oldnitems;
+				oldcap;
 	Datum	   *olditemsarr;
 
-	if (resarr->nitems < resarr->capacity)
+	if (resarr->nitems < resarr->maxitems)
 		return;					/* nothing to do */
 
 	olditemsarr = resarr->itemsarr;
 	oldcap = resarr->capacity;
-	oldnitems = resarr->nitems;
 
 	resarr->capacity = oldcap > 0 ? oldcap * 2 : RESARRAY_INIT_SIZE;
 	resarr->itemsarr = (Datum *)
 		MemoryContextAlloc(TopMemoryContext,
 						   resarr->capacity * sizeof(Datum));
+	resarr->maxitems = RESARRAY_MAX_ITEMS(resarr->capacity);
 	resarr->nitems = 0;
 
+	for (i = 0; i < resarr->capacity; i++)
+		resarr->itemsarr[i] = resarr->invalidval;
+
 	if (olditemsarr != NULL)
 	{
-		for (i = 0; i < oldnitems; i++)
-			ResourceArrayAdd(resarr, olditemsarr[i]);
+		while (oldcap > 0)
+		{
+			oldcap--;
+			if (olditemsarr[oldcap] != resarr->invalidval)
+				ResourceArrayAdd(resarr, olditemsarr[oldcap]);
+		}
 		pfree(olditemsarr);
 	}
 }
@@ -152,12 +202,24 @@ ResourceArrayEnlarge(ResourceArray * resarr)
 static bool
 ResourceArrayGetAny(ResourceArray * resarr, Datum *out)
 {
+	uint32		mask;
+
 	if (resarr->nitems == 0)
 		return false;
 
 	Assert(resarr->capacity > 0);
+	mask = resarr->capacity - 1;
+
+	for (;;)
+	{
+		resarr->lastidx = resarr->lastidx & mask;
+		if (resarr->itemsarr[resarr->lastidx] != resarr->invalidval)
+			break;
+
+		resarr->lastidx++;
+	}
 
-	*out = resarr->itemsarr[resarr->nitems - 1];
+	*out = resarr->itemsarr[resarr->lastidx];
 	return true;
 }
 
@@ -170,6 +232,7 @@ ResourceArrayFree(ResourceArray * resarr)
 	Assert(resarr->nitems == 0);
 
 	resarr->capacity = 0;
+	resarr->maxitems = 0;
 
 	if (!resarr->itemsarr)
 		return;
@@ -287,15 +350,15 @@ ResourceOwnerCreate(ResourceOwner parent, const char *name)
 		parent->firstchild = owner;
 	}
 
-	ResourceArrayInit(&(owner->catrefarr));
-	ResourceArrayInit(&(owner->catlistrefarr));
-	ResourceArrayInit(&(owner->relrefarr));
-	ResourceArrayInit(&(owner->planrefarr));
-	ResourceArrayInit(&(owner->tupdescarr));
-	ResourceArrayInit(&(owner->snapshotarr));
-	ResourceArrayInit(&(owner->dsmarr));
-	ResourceArrayInit(&(owner->bufferarr));
-	ResourceArrayInit(&(owner->filearr));
+	ResourceArrayInit(&(owner->catrefarr), (Datum) (NULL));
+	ResourceArrayInit(&(owner->catlistrefarr), (Datum) (NULL));
+	ResourceArrayInit(&(owner->relrefarr), (Datum) (NULL));
+	ResourceArrayInit(&(owner->planrefarr), (Datum) (NULL));
+	ResourceArrayInit(&(owner->tupdescarr), (Datum) (NULL));
+	ResourceArrayInit(&(owner->snapshotarr), (Datum) (NULL));
+	ResourceArrayInit(&(owner->dsmarr), (Datum) (NULL));
+	ResourceArrayInit(&(owner->bufferarr), (Datum) (InvalidBuffer));
+	ResourceArrayInit(&(owner->filearr), (Datum) (-1));
 
 	return owner;
 }
